/*
 * Copyright (C) 1997-2004 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "server.h"
#include "config.h"
#include <sys/mman.h>
/* #include <time.h> */

#define BACKLOG 10 /*  how many pending connections queue will hold */

/* these are exported to afs/command/dbtool */
struct misc_meta_data *mmd;
struct gengetopt_args_info conf;
char *user_list;

struct dbtool dblist[] = {
	{
		.name = "dopey",
		.init = dopey_dbtool_init,
	},
#if HAVE_MYSQL
	{
		.name = "mysql",
		.init = mysql_dbtool_init,
	},
#endif
	{
		.name = NULL,
	}
};


/* global variables for server-internal use */
static FILE *logfile;
static char *logfile_name;
static int mmd_semid;
static int signal_pipe[2];

/*
 * setup shared memory area and get semaphore for locking
 */
static void shm_init(void)
{
	int fd;
	caddr_t area;

	if ((fd = open("/dev/zero", O_RDWR)) < 0) {
		clog(EMERG, "%s: failed to open /dev/zero\n", __func__);
		exit(EXIT_FAILURE);
	}
	if ((area = mmap(0, sizeof(struct misc_meta_data),
			PROT_READ | PROT_WRITE,
			MAP_SHARED, fd, 0)) == (caddr_t) - 1) {
		clog(EMERG, "%s: mmap error\n", __func__);
		exit(EXIT_FAILURE);
	}
	close(fd); /* we dont need /dev/zero anymore */
	mmd = (struct misc_meta_data *)area;
	
	mmd->dbt_num = 0;
	mmd->num_played = 0;
	mmd->num_commands = 0;
	mmd->num_connects = 0;
	mmd->active_connections = 0;
	strcpy(mmd->filename, "(none)");
	mmd->audio_format = -1;
	mmd->afs_status_flags = AFS_NEXT;

	mmd_semid = semget(42, 1, IPC_CREAT | 0666);
	if (mmd_semid == -1) {
		clog(EMERG, "%s: semget failed\n", __func__);
		exit(EXIT_FAILURE);
	}
}

static void do_semop(struct sembuf *sops, int num)
{
	if (semop(mmd_semid, sops, num) >= 0)
		return;
	clog(WARNING, "%s: semop failed (%s), retrying\n", __func__,
		strerror(errno));
	while (semop(mmd_semid, sops, num) < 0)
		; /* nothing */
}

/*
 * lock shared memory area
 */
void mmd_lock(void)
{
	struct sembuf sops[2] = {
		{
			.sem_num = 0,
			.sem_op = 0,
			.sem_flg = SEM_UNDO
		},
		{
			.sem_num = 0,
			.sem_op = 1,
			.sem_flg = SEM_UNDO
		}
	};
	do_semop(sops, 2);
}

/*
 * unlock shared memory area
 */
void mmd_unlock(void)
{
	struct sembuf sops[1] = {
		{
			.sem_num = 0,
			.sem_op = -1,
			.sem_flg = SEM_UNDO
		},
	};
	do_semop(sops, 1);
}

/*
 * set or get server's uptime
 */
int server_uptime(int set_or_get)
{
	static time_t startuptime;
	time_t now;

	if (set_or_get == SET) {
		time(&startuptime);
		return 0;
	}
	time(&now);
	return((int) difftime(now, startuptime));
}
/*
 * do the usual stuff to become a daemon
 */
static void daemon_init(void)
{
	pid_t pid;
	
	clog(INFO, "%s: daemonizing\n", __func__);
	if ((pid = fork()) < 0) {
		clog(EMERG, "%s: Failed to init daemon\n", __func__);
		exit(EXIT_FAILURE);
	} else if (pid)
		exit(EXIT_SUCCESS); /* parent exits */
	/* child */
	setsid(); /* become session leader */
	chdir("/");
	umask(0);
}

/*
 * die on sigint/sigterm. Kill all children too.
 */
static void sigint_sigterm_handler(int s)
{
	signal(SIGINT, SIG_IGN);
	signal(SIGTERM, SIG_IGN);
	kill(0, SIGTERM);
	exit(s);
}

/*
 * just write one integer to signal pipe
 */
static void generic_signal_handler(int s)
{
	write(signal_pipe[1], &s, sizeof(int));
}

static void handle_sigchld(void)
{
	int status;
	pid_t pid;

	while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
		if (WIFEXITED(status)) {
			clog(DEBUG, "%s: child %i exited. Exit status: %i\n",
				__func__, pid, WEXITSTATUS(status));
			continue;
		}
		if (WIFSIGNALED(status)) {
			clog(DEBUG, "%s: child %i was killed by signal %i\n",
				__func__, pid, WTERMSIG(status));
			continue;
		}
		clog(WARNING, "%s: child %i terminated abormally\n",
			__func__, pid);
	}
}

/*
 * log function. first argument is loglevel.
 */
void clog(int ll, char* fmt,...)
{
	va_list argp;
	FILE *outfd;
	struct tm *tm;
	time_t t1;
	char str[MAXLINE] = "";
	pid_t mypid;

	if (mmd && ll < conf.loglevel_arg)
		return;
	if (!logfile) {
		if (ll < ERROR)
			outfd = stdout;
		else
			outfd = stderr;
	} else
		outfd = logfile;
	time(&t1);
	tm = localtime(&t1);
	strftime(str, MAXLINE, "%b %d %H:%M:%S", tm);
	fprintf(outfd, "%s ", str);	
	if (conf.loglevel_arg <= VERBOSE)
		fprintf(outfd, "%i: ", ll);
	mypid = getpid();
	if (conf.loglevel_arg <= VERBOSE)
		fprintf(outfd, "(%d) ", mypid);
	if (conf.loglevel_arg > VERBOSE) {
		fmt = strchr(fmt, ':');
		if (!fmt)
			return;
		fmt++;
		fmt++;
	}
	va_start(argp, fmt);
	if (conf.loglevel_arg > VERBOSE)
		va_arg(argp, char*);
	vfprintf(outfd, fmt, argp);
	va_end(argp);
}


static void open_log(void)
{
	if (!logfile_name || logfile)
		return;
	if (!(logfile = fopen(logfile_name, "a"))) {
		clog(EMERG, "%s: opening logfile %s failed (%s)", __func__,
			logfile_name, strerror(errno));
		exit(EXIT_FAILURE);
	}
	setlinebuf(logfile);
	clog(INFO, "%s: logfile %s opened\n", __func__, logfile_name);
}

static void close_log(void)
{
	if (!logfile)
		return;
	clog(INFO, "%s: closing logfile\n", __func__);
	fclose(logfile);
	logfile = NULL;
}

static void parse_config(int override)
{
	char *home;
	struct gengetopt_args_info *args_info = &conf;
	static char default_config_file[_POSIX_PATH_MAX];
	static char default_user_list[_POSIX_PATH_MAX];
	struct stat statbuf;
	int ret;
	char *cf = default_config_file;

	if (!args_info->config_file_given || !args_info->user_list_given) {
		if (!(home = getenv("HOME"))) {
			clog(EMERG, "%s: getenv error, cannot read HOME\n",
				__func__);
			exit(EXIT_FAILURE);
		}
	}
	if (args_info->config_file_given)
		cf = args_info->config_file_arg;
	else
		sprintf(cf, "%s/.paraslash/server.conf", home);
	ret = stat(cf, &statbuf);
	if (ret && args_info->config_file_given) {
		fprintf(stderr, "Can't stat config file %s\n", cf);
		exit(EXIT_FAILURE);
	}
	if (!ret)
		cmdline_parser_configfile(cf, args_info, override, 0, 0);
	/* user list */
	if (!args_info->user_list_given) {
		user_list = default_user_list;
		sprintf(user_list, "%s/.paraslash/server.users", home);
	} else
		user_list = args_info->user_list_arg;
	/* logfile */
	if (args_info->logfile_given) {
		logfile_name = args_info->logfile_arg;
		open_log();
	} else {
		if (conf.daemon_given) {
			clog(EMERG, "%s: daemon mode, but no logfile\n",
				__func__);
			exit(EXIT_FAILURE);
		}
		close_log();
		logfile_name = NULL;
		logfile = NULL;
	}
}

/*
 * called when server gets SIGHUP. Reread configuration, then close and
 * reopen log.
 */
static int handle_sighup(void)
{
	clog(NOTICE, "%s: Got SIGHUP\n", __func__);
	close_log(); /* gets reopened if necessary by parse_config */
	parse_config(1);
	return 0;
}

static void install_sighandler(int sig, void (*fct)(int), int restart)
{
	struct sigaction sa;

	clog(DEBUG, "%s: Installing handler for signal %d\n", __func__, sig);
	sa.sa_handler = fct;
	sigemptyset(&(sa.sa_mask));
	if (restart)
		sa.sa_flags = SA_RESTART;
	if (sigaction(sig, &sa, NULL) == -1) {
		clog(EMERG, "%s: Failed to install sighandler %d\n", __func__,
			sig);
		exit(EXIT_FAILURE);
	}
}

static void install_sighandlers(void)
{
	if (pipe(signal_pipe) < 0) {
		clog(ERROR, "%s: %s", __func__, strerror(errno));
		exit(EXIT_FAILURE);
	}
	install_sighandler(SIGINT, &sigint_sigterm_handler, 0);
	install_sighandler(SIGTERM, &sigint_sigterm_handler, 0);
	install_sighandler(SIGHUP, &generic_signal_handler, 0);
	install_sighandler(SIGCHLD, &generic_signal_handler, 1);
	signal(SIGPIPE, SIG_IGN);
}

void init_dbtool()
{
	int i;

	mmd->dbt_change = -1; /* no change nec., set to new dbt num by com_cdt */
	clog(NOTICE, "%s: initializing database tool\n", __func__);
	if (!dblist[1].name)
		goto dopey;
	if (conf.dbtool_given) {
		for (i = 0; dblist[i].name; i++) {
			if (strcmp(dblist[i].name, conf.dbtool_arg))
				continue;
			if (dblist[i].init(&dblist[i]) < 0) {
				clog(WARNING, "%s: init %s failed", 
					dblist[i].name, __func__);
				goto dopey;
			}
			mmd->dbt_num = i;
			return;
		}
		clog(WARNING, "%s: no such dbtool, switching to dopey\n", 
			__func__);
		goto dopey;
	}
	/* use the first dbtool that works 
	 * (assuming that dopey always works) 
	 */
	for (i = 1; dblist[i].name; i++) {
		if (dblist[i].init(&dblist[i]) >= 0) {
			clog(INFO, "%s: initialized %s\n", __func__, 
				dblist[i].name);
			mmd->dbt_num = i;
			return;
		}
		clog(CRIT, "%s: %s init failed\n", __func__, 
			dblist[i].name);
	}
dopey:
		mmd->dbt_num = 0;
		dblist[0].init(&dblist[0]); /* always successful */
}

static int init_network(void)
{
	int sockfd;
	struct sockaddr_in my_addr;

	clog(INFO, "%s: init network socket\n", __func__);
	if ((sockfd = get_socket()) < 0)
		exit(EXIT_FAILURE);
	if (setserversockopts(sockfd) < 0)
		exit(EXIT_FAILURE);
	init_sockaddr(&my_addr, conf.port_arg, NULL);
	if (do_bind(sockfd, &my_addr) < 0)
		exit(EXIT_FAILURE);
	if (listen(sockfd, BACKLOG) == -1) {
		clog(EMERG, "%s: listen failed\n", __func__);
		exit(EXIT_FAILURE);
	}
	return sockfd;
}


static int do_inits(int argc, char **argv)
{
	/* connector's address information */
	int sockfd;
	
	/* init mmd struct */
	shm_init();
	/* parse command line options */
	cmdline_parser(argc, argv, &conf);
	/* parse config file, open log and set defaults */
	parse_config(0);
	/* reset server uptime */
	server_uptime(SET);
	/* print/log welcome message */
	if (conf.loglevel_arg < NOTICE)
		cmdline_parser_print_version();
	clog(NOTICE, "%s: ****** para_server " VERSION " startup ******\n",
		__func__);
	clog(INFO, "%s: build: %s\n", __func__, BUILD_DATE);
	clog(DEBUG, "%s: using loglevel %d\n", __func__, conf.loglevel_arg);
	init_dbtool();
	clog(NOTICE, "%s: initializing audio file sender\n", __func__);
	/* audio file sender */
	if (afs_init(&conf) < 0)
		exit(EXIT_FAILURE);
	/* become daemon */
	if (conf.daemon_given)
		daemon_init();
	mmd->server_pid = getpid();
	/* setup signal handlers */
	clog(NOTICE, "%s: setting up signal handlers\n", __func__);
	install_sighandlers();
	mmd_lock();
	/* init network socket */
	clog(NOTICE, "%s: initializing networking\n", __func__);
	sockfd = init_network();
	if (conf.autoplay_given)
		mmd->afs_status_flags |= AFS_PLAYING;
	clog(NOTICE, "%s: init complete\n", __func__);
	return sockfd;
}

/*
 * taken from libdaemon
 */
static int daemon_signal_next(void)
{
	int s;
	ssize_t r;

	if ((r = read(signal_pipe[0], &s, sizeof(s))) == sizeof(s))
		return s;
	if (r < 0) {
		if (errno == EAGAIN)
			return 0;
		else {
			clog(ERROR, "%s: %s", __func__, strerror(errno));
			return -1;
		}
	}
	clog(ERROR, "%s: Short read() on signal pipe\n", __func__);
	return -1;
}

static void handle_dbt_change(void)
{
	int ret, old = mmd->dbt_num, new = mmd->dbt_change;
	
	dblist[old].shutdown();
	ret = dblist[new].init(&dblist[new]);
	mmd->dbt_change = -1; /* reset */
	if (ret >= 0) {
		mmd->dbt_num = new;
		return;
	}
	/* switch to dopey since init failed */
	dblist[0].init(&dblist[0]);
	mmd->dbt_num = 0; 
}

/*
 * MAIN
 */
int main(int argc, char **argv)
{
	/* listen on sock_fd, new connection on new_fd */
	int sockfd, new_fd;
	struct sockaddr_in their_addr;
	int err, i, sw_fileno, max_fileno, ret, check_stream_writer;
	pid_t chld_pid;
	fd_set rfds, wfds;
	FILE* stream_writer = NULL;

	sockfd = do_inits(argc, argv);
repeat:
	/* check socket and signal pipe in any case */
	FD_ZERO(&rfds);
	FD_SET(sockfd, &rfds);
	max_fileno = sockfd;
	FD_SET(signal_pipe[0], &rfds);
	max_fileno = MAX(max_fileno, signal_pipe[0]);

	check_stream_writer = 0;
	if (mmd->afs_status_flags & AFS_PLAYING) {
		stream_writer = afs_preselect();
		/* do we still want to play? */
		if (stream_writer) {
			check_stream_writer = 1;
			sw_fileno = fileno(stream_writer);
			FD_ZERO(&wfds);
			FD_SET(sw_fileno, &wfds);
			max_fileno = MAX(sw_fileno, max_fileno);
		}
	}
	mmd_unlock();
	//clog(DEBUG, "%s: select (max = %i)\n", __func__, max_fileno);
	ret = select(max_fileno + 1, &rfds, check_stream_writer? &wfds : NULL,
		NULL, NULL);
	err = errno;
	//clog(DEBUG, "%s: select returned  %i\n", __func__, ret);
	mmd_lock();
	if (mmd->dbt_change >= 0)
		handle_dbt_change();
	if (ret < 0 && err == EINTR)
		goto repeat;
	if (ret < 0) {
		clog(CRIT, "%s: select error (%s)\n", __func__, strerror(err));
		goto repeat;
	}
	if (check_stream_writer && FD_ISSET(sw_fileno, &wfds)
			&& mmd->afs_status_flags & AFS_PLAYING)
		afs_send_chunk(stream_writer);
	if (FD_ISSET(signal_pipe[0], &rfds)) {
		int sig_nr;
		sig_nr = daemon_signal_next();
		clog(DEBUG, "%s: caught signal %d\n", __func__, sig_nr);
		switch (sig_nr) {
		case SIGHUP:
			handle_sighup();
			break;
		case SIGCHLD:
			handle_sigchld();
			break;
		}
	}
	if (!FD_ISSET(sockfd, &rfds))
		goto repeat;
	
	new_fd = do_accept(sockfd, &their_addr);
	if (new_fd < 0)
		goto repeat;
	clog(NOTICE, "%s: got connection from %s\n", __func__,
		inet_ntoa(their_addr.sin_addr));
	mmd->num_connects++;
	mmd->active_connections++;
	chld_pid = fork();
	if (chld_pid < 0) {
		clog(ERROR, "%s: fork failed\n", __func__);
		goto repeat;
	}
	if (chld_pid) {
		close(new_fd);
		/* parent keeps accepting connections */
		goto repeat;
	}
	/* child doesn't need the listener and ignores SIGCHLD */
	close(sockfd);
	signal(SIGCHLD, SIG_IGN);
	/*
	 * put info on who we are serving into argv[0] to make
	 * client ip visible in top/ps
	 */
	for (i = argc - 1; i >= 0; i--)
		memset(argv[i], 0, strlen(argv[i]));
	sprintf(argv[0], "para_server (serving %s)",
		inet_ntoa(their_addr.sin_addr));
	handle_connect(new_fd); /* never returns */
	return 0; /* shut up, gcc! */
}
